
class _DataInfo:
    def __init__(self, loader):
        self.cached = None
        self.loader = loader


def _get_data_property(hidden_name):

    def _cached_data(self):
        info = getattr(self, hidden_name)  # type: _DataInfo
        if info.cached is None:
            info.cached = info.loader(*self._args, **self._kwargs)
        return info.cached

    return property(_cached_data)


class DataMeta(type):
    """Meta class for BaseData subclasses"""
    def __new__(mcs, cls_name, bases, attrib):
        mapping = {}

        for name, value in attrib.items():
            if name.startswith('_'):
                mapping[name] = value
                continue

            hidden_name = '_' + name

            mapping[hidden_name] = _DataInfo(value)
            mapping[name] = _get_data_property(hidden_name)

        return super().__new__(mcs, cls_name, bases, mapping)



import io
import logging
import pathlib

import os
import traceback

import sys
from contextlib import redirect_stdout

from .path import Path


def _get_logger():
    return logging.getLogger(__name__)


def _print_loader_exception(e):
    msg = io.StringIO()
    with redirect_stdout(msg):
        traceback.print_exc()

    msg_lines = msg.getvalue().split('\n')
    msg_lines.append('{}: {}'.format(e.__class__.__name__, e))

    for line in msg_lines:
        _get_logger().info('LOADER: {}'.format(line))


def _get_path_property(hidden_name):

    loader_triggered = False

    def _joined_path(self):
        nonlocal loader_triggered

        path = getattr(self, hidden_name)
        abs_path = self._root / str(path)

        if not loader_triggered:
            _get_logger().info('Running loader for {}'.format(path))
            try:
                path.loader(abs_path, workspace=self._workspace)
            except Exception as e:
                _get_logger().warning('Loader for {} failed'.format(path))
                _print_loader_exception(e)
                return None

            _get_logger().info('Finished: loader for {}'.format(path))
            loader_triggered = True

        return abs_path

    return property(_joined_path)


def default_loader(path: pathlib.Path, **kwargs):
    is_dir = '.' not in path.name

    if is_dir:
        os.makedirs(str(path))


class PathStructureMeta(type):
    def __new__(mcs, cls_name, bases, attrib):
        mapping = {}

        for name, value in attrib.items():
            if name.startswith('_'):
                mapping[name] = value
                continue

            hidden_name = '_' + name

            if isinstance(value, Path):
                loader = value.loader
            elif isinstance(value, str) or isinstance(value, pathlib.Path):
                loader = default_loader
            else:
                raise ValueError('PathStructure value type must be one of {}'.format(
                    Path, str, pathlib.Path,
                ))

            mapping[hidden_name] = Path(str(value), loader)
            mapping[name] = _get_path_property(hidden_name)

        return super().__new__(mcs, cls_name, bases, mapping)



import os
from typing import Type

from .base_path_structure import BasePathStructure
from .action_tracker import ActionTracker
from .base_action import BaseAction
from .state import WorkspaceState
from .base_data import BaseData


class Workspace:
    """Base class for Workspace subclasses

    Examples
    --------

    :ref:`basic-workspace`

    :ref:`workspace-with-chained-loaders`

    """

    def __init__(self,
                 root,
                 path_structure_cls: Type[BasePathStructure],
                 data_cls: Type[BaseData]=None):
        """Creates a workspace

        Parameters
        ----------
        path_structure_cls: BasePathStructure
            Filled path structure to create Workspace instances
        data_cls: BaseData
            Interface to automatically access data in the workspace
        """
        self.path = path_structure_cls(root, workspace=self)
        self.state_path = self.root / '.state.json'
        self.state = WorkspaceState(self.state_path)
        self.data = data_cls(self) if data_cls else None

    @property
    def root(self):
        return self.path._root

    def __repr__(self):
        return '<{}({})>'.format(self.__class__.__name__, self.root)

    @property
    def id(self):
        """Workspace id attribute

        Id is used to sort workspaces from oldest to latest
        """
        dirname = self.root.name
        try:
            return int(dirname)
        except ValueError:
            return str(dirname)

    def track(self, action: BaseAction) -> ActionTracker:
        """Tracks the action inside the workspace

        Creates an ActionTracker object with action and this workspace as constructor parameters

        Parameters
        ----------
        action: subclass of BaseAction
            Action to track
        """
        return ActionTracker(action, self)

    def __eq__(self, other):
        return self.root == other.root



import json
import os


def _store_state(state, path):
    with open(str(path), 'w') as f:
        json.dump(state, f, indent=4)
        f.write('\n')


def _load_state(path):
    with open(str(path), 'r') as f:
        state = json.load(f)
    return state


class WorkspaceState:
    """Manages workspace state

    State is written to file each time anything changes
    """

    def __init__(self, state_path):
        self.state_path = state_path

        if not os.path.exists(str(self.state_path)):
            state = {}
            _store_state(state, self.state_path)

    def __getitem__(self, item):
        state = _load_state(self.state_path)
        return state[item] if item in state else None

    def __setitem__(self, key, value):
        path = self.state_path
        state = _load_state(path)
        state[key] = value
        _store_state(state, path)

    def __iter__(self):
        return iter(_load_state(self.state_path).items())



import os
import warnings
from pathlib import Path
from typing import Optional, Type

from .base_path_structure import BasePathStructure
from .state import WorkspaceState
from .storage import WorkspaceStorage
from .workspace import Workspace
from .base_data import BaseData


_TARGET = 'target'
_CURRENT = 'current'
_STATE = '.state.json'


class WorkspaceManager:
    """Manages workspaces

    Allows you to
    - detect workspaces
    - manage current, target and latest workspaces
    - track workspace actions
    - search for workspaces with finished actions

    Exmaples
    --------

    :ref:`workspace-management`

    :ref:`manager-with-data-bindings`
    """

    def __init__(self,
                 path,
                 path_structure_cls: Type[BasePathStructure],
                 data_cls: Type[BaseData]=None,
                 *,
                 workspace_cls=None):
        """Initialize workspace manager

        Parameters
        ----------
        path: path-like
            Workspace storage path
        workspace_cls: class object of BaseWorkspace subclass
            Workspace class object
        """

        self.workspace_cls = (
            workspace_cls
            if workspace_cls
            else Workspace
        )
        self.root = Path(path)

        os.makedirs(str(self.root), exist_ok=True)

        self.path_structure_cls = path_structure_cls
        self.data_cls = data_cls

        self.storage = WorkspaceStorage.load_from_directory(
            self.root,
            self.path_structure_cls,
            data_cls=self.data_cls,
            workspace_cls=self.workspace_cls,
        )

        self.state = WorkspaceState(self._state_path)

    @property
    def current_ws_id(self):
        return self.state[_CURRENT]

    @current_ws_id.setter
    def current_ws_id(self, value):
        self.state[_CURRENT] = value

    @property
    def target_ws_id(self):
        return self.state[_TARGET]

    @target_ws_id.setter
    def target_ws_id(self, value):
        self.state[_TARGET] = value

    @property
    def _state_path(self):
        """Path to the state file

        Is used internally
        """
        return self.root / _STATE

    def create(self, name) -> Workspace:
        """Creates a Workspace

        Parameters
        ----------
        name: str or int
            Workspace name
        """
        if isinstance(name, int):
            name = str(name)

        ws_root = self.root / name
        ws = self.workspace_cls(ws_root, self.path_structure_cls, self.data_cls)
        self.storage.add(ws)
        return ws

    def latest(self) -> Workspace:
        """Returns latest workspace"""
        return self.storage.at(-1)

    def current(self) -> Optional[Workspace]:
        """Returns current workspace"""
        return self.storage[self.current_ws_id]

    def target(self) -> Workspace:
        """Returns target workspace"""
        return self.storage[self.target_ws_id]

    def __repr__(self):
        return '<{}(root={}, storage={})>'.format(
            self.__class__.__name__,
            self.root,
            self.storage,
        )

    def find_latest_finished(self, action) -> Optional[Workspace]:
        """Searches for latest workspace with finished action constraint"""
        workspaces = list(self.storage.values())
        for ws in reversed(workspaces):
            tracker = ws.track(action)
            if tracker.finished():
                return ws

    def target_latest(self):
        self.target_ws_id = self.latest().id

    def update(self):
        """Move from current workspace to target one"""
        if self.current_ws_id == self.target_ws_id:
            warnings.warn('Current and target workspaces are the same')
            return

        self.current_ws_id = self.target_ws_id
